##
# File:    ArchiveIoBase.py
# Author:  jdw
# Date:    10-Oct-2017
# Version: 0.001
#
# Updates:
#
##
"""
Base class for archive data transfer operation utilities.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.001"

#
#
import logging

from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId

logger = logging.getLogger(__name__)


class ArchiveIoBase(object):
    """A base class for for archive data transfer operation utilities."""

    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        self._raiseExceptions = kwargs.get("raiseExceptions", False)
        self._siteId = kwargs.get("siteId", getSiteId())
        self._serverId = kwargs.get("serverId", None)

        self.__cI = ConfigInfo(siteId=getSiteId())
        #
        cD = self.__cI.get(self._serverId, {})
        self._hostName = cD.get("HOST_NAME", None)
        self._userName = cD.get("HOST_USERNAME", None)
        self._password = cD.get("HOST_PASSWORD", None)
        self._hostPort = int(cD.get("HOST_PORT", 22))
        self._protocol = cD.get("HOST_PROTOCOL", None)
        self._rootPath = cD.get("HOST_ROOT_PATH", None)
        self._keyFilePath = cD.get("HOST_KEY_FILE_PATH", None)
        self._keyFileType = cD.get("HOST_KEY_FILE_TYPE", None)
        #

    def connect(self, hostName, userName, **kwargs):
        raise NotImplementedError("To be implemented in subclass")

    def mkdir(self, path, **kwargs):
        raise NotImplementedError("To be implemented in subclass")

    def stat(self, path):
        raise NotImplementedError("To be implemented in subclass")

    def put(self, localPath, remotePath):
        raise NotImplementedError("To be implemented in subclass")

    def get(self, remotePath, localPath):
        raise NotImplementedError("To be implemented in subclass")

    def listdir(self, path):
        raise NotImplementedError("To be implemented in subclass")

    def rmdir(self, path):
        raise NotImplementedError("To be implemented in subclass")

    def remove(self, path):
        raise NotImplementedError("To be implemented in subclass")

    def close(self):
        raise NotImplementedError("To be implemented in subclass")



##
# File: CvsUtility.py
# Date: 09-April-2011  j. Westbrook
#
# Updates:
# 12-April-2011 jdw Revision checkout
##
"""
Wrapper class for opertations on cvs repositories.

Methods are provided to manage archiving of chemical component data files.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.001"


import os
import subprocess
import tempfile
import shutil
import logging


class CvsWrapper(object):
    """Wrapper class for opertations on cvs repositories."""

    def __init__(self, tmpPath="./"):
        self.__tmpPath = tmpPath
        self.__logger = logging.getLogger("wwpdb.utils.rcsb.CvsWrapper")
        self.__logger.debug("Created instance of CvsWrapper")
        #
        # self.__debug = True
        self.__repositoryHost = None
        self.__repositoryPath = None
        self.__cvsUser = None
        self.__cvsPassword = None
        self.__cvsRoot = None
        #
        self.__wrkPath = None
        self.__cvsInfoFileName = "cvsInfo.txt"
        self.__cvsErrorFileName = "cvsError.txt"

    def setRepositoryPath(self, host, path):
        self.__repositoryHost = host
        self.__repositoryPath = path

    def setAuthInfo(self, user, password):
        self.__cvsUser = user
        self.__cvsPassword = password

    def getHistory(self, cvsPath):
        text = ""
        cmd = self.__getHistoryCmd(cvsPath)
        if cmd is not None:
            _ok = self.__runCvsCommand(myCommand=cmd)  # noqa: F841
            text = self.__getOutputText()

        return text

    def getRevisionList(self, cvsPath):
        """Return a list of tuples containing the revision identifiers for the input file.

        Return data has the for [(RevId, A/M, timeStamp),...] where A=Added and M=Modified.
        """
        revList = []
        cmd = self.__getHistoryCmd(cvsPath)
        if cmd is not None:
            _ok = self.__runCvsCommand(myCommand=cmd)  # noqa: F841
            revList = self.__extractRevisions()
        return revList

    def cleanup(self):
        """Cleanup temporary files and directories"""
        return shutil.rmtree(self.__wrkPath)

    def checkOutFile(self, cvsPath, outPath, revId=None):
        text = ""
        (pth, fn) = os.path.split(cvsPath)
        self.__logger.debug("Cvs directory %s   target file name %s", pth, fn)
        if len(fn) > 0:
            cmd = self.__getCheckOutCmd(cvsPath, outPath, revId)
            if cmd is not None:
                _ok = self.__runCvsCommand(myCommand=cmd)  # noqa: F841
                text = self.__getErrorText()
        else:
            pass
        return text

    def __getHistoryCmd(self, cvsPath):
        if self.__wrkPath is None:
            self.__makeTempWorkingDir()
        outPath = os.path.join(self.__wrkPath, self.__cvsInfoFileName)
        errPath = os.path.join(self.__wrkPath, self.__cvsErrorFileName)
        if self.__setCvsRoot():
            cmd = "cvs -d " + self.__cvsRoot + " history -a -x AM " + cvsPath + self.__getRedirect(fileNameOut=outPath, fileNameErr=errPath)
        else:
            cmd = None
        return cmd

    def __getCheckOutCmd(self, cvsPath, outPath, revId=None):
        if self.__wrkPath is None:
            self.__makeTempWorkingDir()
        errPath = os.path.join(self.__wrkPath, self.__cvsErrorFileName)
        (pth, fn) = os.path.split(cvsPath)
        self.__logger.debug("CVS directory %s  target file name %s", pth, fn)
        lclPath = os.path.join(self.__wrkPath, fn)
        #
        #
        if self.__setCvsRoot():
            if revId is None:
                rS = " "
                cmd = (
                    "cvs -d "
                    + self.__cvsRoot
                    + " co -d "
                    + self.__wrkPath
                    + " "
                    + cvsPath
                    + self.__getRedirect(fileNameOut=errPath, fileNameErr=errPath)
                    + " ; "
                    + " mv -f  "
                    + lclPath
                    + " "
                    + outPath
                    + self.__getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True)
                    + " ; "
                )
            else:
                rS = " -r " + revId + " "
                cmd = (
                    "cvs -d "
                    + self.__cvsRoot
                    + " co -d "
                    + self.__wrkPath
                    + rS
                    + " "
                    + cvsPath
                    + self.__getRedirect(fileNameOut=errPath, fileNameErr=errPath)
                    + " ; "
                    + " mv -f  "
                    + lclPath
                    + " "
                    + outPath
                    + self.__getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True)
                    + " ; "
                )
        else:
            cmd = None
        return cmd

    def __getRedirect(self, fileNameOut="myLog.log", fileNameErr="myLog.log", append=False):
        if append:
            if fileNameOut == fileNameErr:
                oReDir = " >> " + fileNameOut + " 2>&1 "
            else:
                oReDir = " >> " + fileNameOut + " 2>> " + fileNameErr
        else:
            if fileNameOut == fileNameErr:
                oReDir = " > " + fileNameOut + " 2>&1 "
            else:
                oReDir = " > " + fileNameOut + " 2> " + fileNameErr

        return oReDir

    def __runCvsCommand(self, myCommand):
        retcode = -100
        try:
            self.__logger.debug("Command: %s", myCommand)

            # if (self.__debug):
            #    self.__lfh.write("+CvsWrapper(__runCvsCommand) command: %s\n" % myCommand)

            retcode = subprocess.call(myCommand, shell=True)
            if retcode < 0:
                self.__logger.debug("Child was terminated by signal %r", retcode)
                # if self.__verbose:
                #    self.__lfh.write("+CvsWrapper(__runCvsCommand) Child was terminated by signal %r\n" % retcode)
                return False
            else:
                self.__logger.debug("Child was terminated by signal %r", retcode)
                # if self.__verbose:
                #    self.__lfh.write("+CvsWrapper(__runCvsCommand) Child was terminated by signal %r\n" % retcode)
                return True
        except OSError as e:
            self.__logger.exception("cvs command exception: %r %r", retcode, str(e))
            # if self.__verbose:
            #    self.__lfh.write("+CvsWrapper(__runCvsCommand) Execution failed: %r\n" % e)
            return False

    def __setCvsRoot(self):
        try:
            self.__cvsRoot = ":pserver:" + self.__cvsUser + ":" + self.__cvsPassword + "@" + self.__repositoryHost + ":" + self.__repositoryPath
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def __extractRevisions(self):
        """Extract revisions details from the last history command."""
        revList = []
        try:
            fName = os.path.join(self.__wrkPath, self.__cvsInfoFileName)
            self.__logger.debug("Reading revisions from %r", fName)
            ifh = open(fName, "r")
            for line in ifh.readlines():
                fields = line[:-1].split()
                typeCode = str(fields[0])
                revId = str(fields[5])
                timeStamp = str(fields[1] + ":" + fields[2])
                revList.append((revId, typeCode, timeStamp))
        except:  # noqa: E722 pylint: disable=bare-except
            self.__logger.exception("Extracting revision list for : %s", fName)

        revList.reverse()
        self.__logger.debug("Ordered revision list %r", revList)

        return revList

    def __getOutputText(self):
        text = ""
        try:
            fPath = os.path.join(self.__wrkPath, self.__cvsInfoFileName)
            ifh = open(fPath, "r")
            text = ifh.read()
        except:  # noqa: E722 pylint: disable=bare-except
            self.__logger.exception("Execption reading cvs output file: %s", fPath)

        return text

    def __getErrorText(self):
        text = ""
        try:
            fName = os.path.join(self.__wrkPath, self.__cvsErrorFileName)
            ifh = open(fName, "r")
            text = ifh.read()
        except:  # noqa: E722 pylint: disable=bare-except
            pass

        return text

    def __makeTempWorkingDir(self):
        if self.__tmpPath is not None and os.path.isdir(self.__tmpPath):
            self.__wrkPath = tempfile.mkdtemp("tmpdir", "rcsbCVS", self.__tmpPath)
        else:
            self.__wrkPath = tempfile.mkdtemp("tmpdir", "rcsbCVS")
        self.__logger.debug("Working directory path set to  %r", self.__wrkPath)



##
# File:  ChemRefPathInfo.py
# Date:  30-Jan-2013
#
# Updated:
# 4-Feb-2013 jdw methods to support testing with surrogate repositories.
# 5-Jul-2013 jdw renamed to ChemRefPathInfo() to avoid name conflict
# 12-Mar-2022 ep Changed API from reqObj to siteId
"""
Common methods for finding path information for chemical reference data files.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os.path

import os
import sys


from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCc


class ChemRefPathInfo(object):
    """Common methods for finding path information for chemical reference data files."""

    #

    def __init__(
        self,
        siteId=None,
        verbose=False,
        log=sys.stderr,
        testMode=False,
        # Old API - to be deprecated
        reqObj=None,
        configObj=None,
        configCommonObj=None,
    ):  # pylint: disable=unused-argument
        """Input request object and configuration (ConfigInfo()) object are used to
        supply information required to compute path details for chemical reference
        data files.
        """
        self.__verbose = verbose
        self.__lfh = log
        if configObj and configCommonObj:
            self.__cI = configObj
            self.__cIcommonAppCc = configCommonObj
        else:
            self.__siteId = siteId
            self.__cI = ConfigInfo(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            self.__cIcommonAppCc = ConfigInfoAppCc(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
        self.__testMode = testMode
        #

    def getIdType(self, idCode):
        id_u = idCode.upper()
        #
        if (idCode is None) or (len(idCode) < 1):
            return None
        if len(id_u) <= 5:
            id_type = "CC"
        elif id_u[:6] == "PRDCC_":
            id_type = "PRDCC"
        elif id_u[:4] == "PRD_":
            id_type = "PRD"
        elif id_u[:4] == "FAM_":
            id_type = "PRD_FAMILY"
        else:
            id_type = None

        return id_type

    def getCcdHash(self, idCode):
        """Returns the hash code for a CCD id.  Currently first letter"""
        if not idCode:
            return None

        if len(idCode) > 3:
            hash_key = idCode.upper()[-2:]
        else:
            hash_key = idCode.upper()[0]

        return hash_key

    def getFilePath(self, idCode, id_type=None):
        """Return the repository file path corresponding to the input reference data id code
        (CC,PRD,FAMILY or PRDCC).

        If id_type is not specified, infer it from length, etc.
        """
        #
        if id_type is None:
            id_type = self.getIdType(idCode)
        if id_type is None:
            return None
        #
        id_u = idCode.upper()
        hash_key = id_u[-1]
        file_name = id_u + ".cif"
        #
        if id_type == "CC":
            hash_key = self.getCcdHash(id_u)
            file_path = os.path.join(self.__cIcommonAppCc.get_site_cc_cvs_path(), hash_key, id_u, file_name)
        elif id_type == "PRDCC":
            file_path = os.path.join(self.__cIcommonAppCc.get_site_prdcc_cvs_path(), hash_key, file_name)
        elif id_type == "PRD":
            file_path = os.path.join(self.__cIcommonAppCc.get_site_prd_cvs_path(), hash_key, file_name)
        elif id_type == "PRD_FAMILY":
            file_path = os.path.join(self.__cIcommonAppCc.get_site_family_cvs_path(), hash_key, file_name)
        else:
            file_path = None

        return file_path

    def getFileDir(self, idCode, id_type=None):
        """Return the repository file directory corresponding to the input reference data id code
        (CC,PRD,FAMILY or PRDCC).

        If id_type is not specified, infer it from length, etc.
        """
        #
        filePath = self.getFilePath(idCode, id_type)
        if filePath:
            return os.path.dirname(filePath)
        return None

    def getProjectPath(self, idCode=None, id_type=None):
        """
        Return the project path for an input reference data id code
        (CC,PRD,FAMILY or PRDCC).
        """
        #
        if id_type is None:
            id_type = self.getIdType(idCode)
        if id_type is None:
            return None
        #
        if id_type == "CC":
            return self.__cIcommonAppCc.get_site_cc_cvs_path()
        elif id_type == "PRDCC":
            return self.__cIcommonAppCc.get_site_prdcc_cvs_path()
        elif id_type == "PRD":
            return self.__cIcommonAppCc.get_site_prd_cvs_path()
        elif id_type == "PRD_FAMILY":
            return self.__cIcommonAppCc.get_site_family_cvs_path()
        else:
            return None

    def getCvsProjectInfo(self, idCode, id_type=None):
        """Assign the CVS project name and relative path based on the input ID code.

        The project name represents the directory containing the checked out
        repository within the sandbox directory.  Relative path identifies gives
        the target path for the target file within the repository.
        """
        rel_path = None
        project_name = None

        if id_type is None:
            id_type = self.getIdType(idCode)
        if id_type is None:
            return project_name, rel_path
        #
        project_name = self.assignCvsProjectName(id_type)
        if id_type == "CC":
            hash_key = self.getCcdHash(idCode)
            rel_path = os.path.join(hash_key, idCode, idCode + ".cif")
        elif id_type == "PRDCC":
            rel_path = os.path.join(idCode[-1], idCode + ".cif")
        elif id_type == "PRD":
            rel_path = os.path.join(idCode[-1], idCode + ".cif")
        elif id_type == "PRD_FAMILY":
            rel_path = os.path.join(idCode[-1], idCode + ".cif")
        else:
            pass

        return project_name, rel_path

    def assignIdCodeFromFileName(self, filePath):
        if self.__verbose:
            self.__lfh.write("+PathInfo.assignIdCodeFromFileName() input file path: %s\n" % filePath)

        if (filePath is not None) and (len(filePath) > 7):
            file_name = os.path.basename(filePath)
            definition_id, _ = os.path.splitext(file_name)
            return definition_id.upper()

        return None

    def assignCvsProjectName(self, repType):
        """Assign the CVS project name from the input repository type.

        This wrapper provides for a testing mode which assign an existing  surogate
        repository project with the same organization as the requested repository type.
        """
        project_name = None

        if repType is None:
            return project_name
        #
        if self.__testMode:
            if repType == "CC":
                project_name = "test-ligand-v1"
            elif repType == "PRDCC":
                project_name = "test-project-v1"
            elif repType == "PRD":
                project_name = "test-project-v1"
            elif repType == "PRD_FAMILY":
                project_name = "test-project-v1"
            else:
                pass
        else:
            if repType == "CC":
                project_name = self.__cI.get("SITE_REFDATA_PROJ_NAME_CC")
            elif repType == "PRDCC":
                project_name = self.__cI.get("SITE_REFDATA_PROJ_NAME_PRDCC")
            elif repType == "PRD":
                project_name = self.__cI.get("SITE_REFDATA_PROJ_NAME_PRD")
            elif repType == "PRD_FAMILY":
                project_name = self.__cI.get("SITE_REFDATA_PROJ_NAME_PRD_FAMILY")
            else:
                pass

        return project_name



##
# File:  PathInfo.py
# Date:  25-Feb-2013  J. Westbrook
#
# Updated:
#  26-Feb-2013  jdw   implement new session storage options in DataFileReference()
#  28-Feb-2013  jdw   move to generic path wwpdb.utils.rcsb.PathInfo()
#  28-Feb-2013  jdw   Add wrappers for more general purpose methods -
#  04-Apr-2013  jdw   Add assembly assignment and map convenience methods
#  27-Aug-2013  jdw   Add optional parameters for content milestone variants [upload,deposit,annotate,...]
#  23-Dec-2013  jdw   Add support for file source 'session-download' as extension of file source session.
#  19-Apr-2014  jdw   add getPolyLinkReportFilePath()
#   9-May-2014  jdw   add entity/partition argument to getSequenceAlignFilePath()
# 25-Jun-2014   jdw   add convenience methods getEmVolumeFilePath() getEmMaskFilePath()
# 26-Jun-2014   jdw   correct parameter errors on blast-match and map methods, add omit-map methods
# 28-Jun-2014   jdw   add template methods for searching file versions and partitions
#  1-Jul-2014   jdw   fix initialization in __getPathWorker()
#  5-Jul-2014   jdw   add method getFilePathContentTypeTemplate()
#  7-Jul-2014   jdw   add method getStatusHistoryFilePath()
# 23-Aug-2014   jdw   add method getEmDepositVolumeParamsFilePath()
# 14-Sep-2014   jdw   add isValidFileName(fileName, requireVersion=True) and splitFileName(fileName)
# 24-Sep-2014   jdw   add getFileExtension(formatType)
# 13-Dec-2016   jdw   add getStructureFactorsPdbxFilePath()
# 23-Oct-2017   jdw   config for logging - add parseFileName wrapper method
# 24-Mar-2018   ep    add mileStone argument to getFilePathContentTypeTemplate()
##
"""
Common methods for finding path information for resource and data files in the wwPDB data processing
and annotation system.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"
__version__ = "V0.07"

import os
import os.path
import logging

from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId
from wwpdb.io.locator.DataReference import DataFileReference, ReferenceFileComponents

logger = logging.getLogger(__name__)


class PathInfo(object):

    """Common methods for finding path information for sequence resources and data files.

    In these methods the parameter contentType refers to a base content type.

    The mileStone parameter is used to select the milestone variant in any convenience methods
    (e.g. model-deposit, model-upload, ... )

    """

    def __init__(self, siteId=None, sessionPath=".", verbose=False, log=None):
        """"""
        self.__verbose = verbose
        self.__lfh = log
        #
        self.__debug = False  # pylint: disable=unused-private-member
        self.__siteId = siteId if siteId is not None else getSiteId(defaultSiteId=siteId)
        self.__sessionPath = sessionPath
        self.__sessionDownloadPath = None
        if self.__sessionPath is not None:
            self.__sessionDownloadPath = os.path.join(self.__sessionPath, "downloads")
        #
        self.__cI = ConfigInfo(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)

    def setDebugFlag(self, flag):
        self.__debug = flag  # pylint: disable=unused-private-member

    def parseFileName(self, fileName):
        rfc = ReferenceFileComponents(verbose=self.__verbose, log=self.__lfh)
        if rfc.set(fileName=fileName):
            return rfc.get()
        else:
            return None, None, None, None, None

    def isValidFileName(self, fileName, requireVersion=True):
        """Is the input file name project compliant ?"""
        rfc = ReferenceFileComponents(verbose=self.__verbose, log=self.__lfh)
        if rfc.set(fileName=fileName):
            (dId, cT, cF, pN, vN) = rfc.get()
            if requireVersion:
                if (dId is None) or (cT is None) or (cF is None) or (pN is None) or (vN is None):
                    return False
                else:
                    return True
            else:
                if (dId is None) or (cT is None) or (cF is None) or (pN is None):
                    return False
                else:
                    return True
        else:
            return False

    def getFileExtension(self, formatType):
        eD = self.__cI.get("FILE_FORMAT_EXTENSION_DICTIONARY")
        try:
            return eD[formatType]
        except Exception as _e:  # noqa: F841
            return None

    def splitFileName(self, fileName):
        """
        returns (depositionDataSetId, contentType, contentFormat, filePartionNumber, [versionId (int) or None])
        """
        try:
            rfc = ReferenceFileComponents(verbose=self.__verbose, log=self.__lfh)
            rfc.set(fileName=fileName)
            return rfc.get()
        except Exception as _e:  # noqa: F841
            return (None, None, None, None, None)

    #

    def setSessionPath(self, sessionPath):
        """Set the top path that will be searched for files with fileSource='session'"""
        self.__sessionPath = sessionPath
        self.__sessionDownloadPath = os.path.join(self.__sessionPath, "downloads")

    def getArchivePath(self, dataSetId):
        try:
            if dataSetId.startswith("G_"):
                return self.getDirPath(dataSetId=dataSetId, fileSource="autogroup")
                # return os.path.join(self.__cI.get('SITE_ARCHIVE_STORAGE_PATH'), 'autogroup', dataSetId)
            else:
                return self.getDirPath(dataSetId=dataSetId, fileSource="archive")
                # return os.path.join(self.__cI.get('SITE_ARCHIVE_STORAGE_PATH'), 'archive', dataSetId)
            #
        except Exception as _e:  # noqa: F841
            return None

    def getInstancePath(self, dataSetId, wfInstanceId):
        try:
            return self.getDirPath(dataSetId=dataSetId, fileSource="wf-instance", wfInstanceId=wfInstanceId)
            # return os.path.join(self.__cI.get('SITE_ARCHIVE_STORAGE_PATH'), 'workflow', dataSetId, 'instance', wfInstanceId)
        except Exception as _e:  # noqa: F841
            return None

    def getInstanceTopPath(self, dataSetId):
        try:
            return os.path.dirname(self.getDirPath(dataSetId=dataSetId, fileSource="wf-instance", wfInstanceId="W_001"))
            # return os.path.join(self.__cI.get('SITE_ARCHIVE_STORAGE_PATH'), 'workflow', dataSetId, 'instance')
        except Exception as _e:  # noqa: F841
            return None

    def getDepositPath(self, dataSetId):
        try:
            return self.getDirPath(dataSetId=dataSetId, fileSource="deposit")
            # return os.path.join(self.__cI.get('SITE_ARCHIVE_STORAGE_PATH'), 'deposit', dataSetId)
        except Exception as _e:  # noqa: F841
            return None

    def getTempDepPath(self, dataSetId):
        try:
            return self.getDirPath(dataSetId=dataSetId, fileSource="tempdep")
        except Exception as _e:  # noqa: F841
            return None

    def getModelPdbxFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, versionId=versionId, contentTypeBase="model", formatType="pdbx", mileStone=mileStone
        )

    def getModelPdbFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, versionId=versionId, contentTypeBase="model", formatType="pdb", mileStone=mileStone
        )

    def getStructureFactorsPdbxFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, versionId=versionId, contentTypeBase="structure-factors", formatType="pdbx", mileStone=mileStone
        )

    def getPolyLinkFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            contentTypeBase="polymer-linkage-distances",
            formatType="pdbx",
            mileStone=mileStone,
        )

    def getPolyLinkReportFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            contentTypeBase="polymer-linkage-report",
            formatType="html",
            mileStone=mileStone,
        )

    def getSequenceStatsFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, versionId=versionId, contentTypeBase="seq-data-stats", formatType="pic", mileStone=mileStone
        )

    def getSequenceAlignFilePath(self, dataSetId, entityId="1", wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            contentTypeBase="seq-align-data",
            partNumber=entityId,
            formatType="pic",
            mileStone=mileStone,
        )

    def getReferenceSequenceFilePath(self, dataSetId, entityId="1", wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber=entityId,
            contentTypeBase="seqdb-match",
            formatType="pdbx",
            mileStone=mileStone,
        )

    def getSequenceAssignmentFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, versionId=versionId, contentTypeBase="seq-assign", formatType="pdbx", mileStone=mileStone
        )

    def getAssemblyAssignmentFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, versionId=versionId, contentTypeBase="assembly-assign", formatType="pdbx", mileStone=mileStone
        )

    def getBlastMatchFilePath(self, dataSetId, entityId="1", wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber=entityId,
            contentTypeBase="blast-match",
            formatType="xml",
            mileStone=mileStone,
        )

    def getMap2fofcFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber="1",
            contentTypeBase="map-2fofc",
            formatType="map",
            mileStone=mileStone,
        )

    def getMapfofcFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber="1",
            contentTypeBase="map-fofc",
            formatType="map",
            mileStone=mileStone,
        )

    def getOmitMap2fofcFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber="1",
            contentTypeBase="omit-map-2fofc",
            formatType="map",
            mileStone=mileStone,
        )

    def getOmitMapfofcFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber="1",
            contentTypeBase="omit-map-fofc",
            formatType="map",
            mileStone=mileStone,
        )

    #

    def getEmVolumeFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber="1",
            contentTypeBase="em-volume",
            formatType="map",
            mileStone=mileStone,
        )

    def getEmMaskFilePath(self, dataSetId, maskNumber="1", wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber=maskNumber,
            contentTypeBase="em-mask",
            formatType="map",
            mileStone=mileStone,
        )

    def getEmDepositVolumeParamsFilePath(self, dataSetId, maskNumber="1", wfInstanceId=None, fileSource="deposit", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber=maskNumber,
            contentTypeBase="deposit-volume-params",
            formatType="pic",
            mileStone=mileStone,
        )

    def getAuthChemcialShiftsFilePath(self, dataSetId, formatType="nmr-star", partNumber="next", wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber=partNumber,
            contentTypeBase="nmr-chemical-shifts-auth",
            formatType=formatType,
            mileStone=mileStone,
        )

    def getChemcialShiftsFilePath(self, dataSetId, formatType="nmr-star", wfInstanceId=None, fileSource="archive", versionId="latest", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource=fileSource,
            versionId=versionId,
            partNumber="1",
            contentTypeBase="nmr-chemical-shifts",
            formatType=formatType,
            mileStone=mileStone,
        )

    def getStatusHistoryFilePath(self, dataSetId, fileSource="archive", versionId="latest"):
        return self.__getStandardPath(
            dataSetId=dataSetId, wfInstanceId=None, fileSource=fileSource, versionId=versionId, partNumber="1", contentTypeBase="status-history", formatType="pdbx", mileStone=None
        )

    #
    def getFilePath(self, dataSetId, wfInstanceId=None, contentType=None, formatType=None, fileSource="archive", versionId="latest", partNumber="1", mileStone=None):
        return self.__getStandardPath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            contentTypeBase=contentType,
            formatType=formatType,
            fileSource=fileSource,
            versionId=versionId,
            partNumber=partNumber,
            mileStone=mileStone,
        )

    #
    def getFileName(self, dataSetId, wfInstanceId=None, contentType=None, formatType=None, fileSource="archive", versionId="latest", partNumber="1", mileStone=None):
        return os.path.basename(
            self.__getStandardPath(
                dataSetId=dataSetId,
                wfInstanceId=wfInstanceId,
                contentTypeBase=contentType,
                formatType=formatType,
                fileSource=fileSource,
                versionId=versionId,
                partNumber=partNumber,
                mileStone=mileStone,
            )
        )

    def getDirPath(
        self, dataSetId, wfInstanceId=None, contentType=None, formatType=None, fileSource="archive", versionId="latest", partNumber="1", mileStone=None
    ):  # noqa: E501 pylint: disable=unused-argument
        dfRef = DataFileReference(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
        dfRef.setDepositionDataSetId(dataSetId)
        dfRef.setStorageType(fileSource)
        if fileSource in ("session", "wf-session"):
            dfRef.setStorageType("session")
            dfRef.setSessionPath(self.__sessionPath)
        if fileSource in ("session-download"):
            dfRef.setStorageType("session")
            dfRef.setSessionPath(self.__sessionDownloadPath)
        dfRef.setSessionDataSetId(dataSetId)
        dfRef.setWorkflowInstanceId(wfInstanceId)
        return dfRef.getDirPathReference()

        # return os.path.dirname(self.__getStandardPath(dataSetId=dataSetId,
        #                                               wfInstanceId=wfInstanceId,
        #                                               contentTypeBase=contentType,
        #                                               formatType=formatType,
        #                                               fileSource=fileSource,
        #                                               versionId=versionId,
        #                                               partNumber=partNumber,
        #                                               mileStone=mileStone))

    def getWebDownloadPath(self, dataSetId, wfInstanceId=None, contentType=None, formatType=None, versionId="latest", partNumber="1", mileStone=None):
        fn = os.path.basename(
            self.__getStandardPath(
                dataSetId=dataSetId,
                wfInstanceId=wfInstanceId,
                contentTypeBase=contentType,
                formatType=formatType,
                fileSource="session-download",
                versionId=versionId,
                partNumber=partNumber,
                mileStone=mileStone,
            )
        )
        (_p, sId) = os.path.split(self.__sessionPath)
        return os.path.join("/sessions", sId, "downloads", fn)

    def getFilePathVersionTemplate(self, dataSetId, wfInstanceId=None, contentType=None, formatType=None, fileSource="archive", partNumber="1", mileStone=None):
        _fp, vt, _pt, _cct = self.__getPathWorker(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            contentTypeBase=contentType,
            formatType=formatType,
            fileSource=fileSource,
            versionId="none",
            partNumber=partNumber,
            mileStone=mileStone,
        )
        return vt

    def getFilePathPartitionTemplate(self, dataSetId, wfInstanceId=None, contentType=None, formatType=None, fileSource="archive", mileStone=None):
        _fp, _vt, pt, _cct = self.__getPathWorker(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            contentTypeBase=contentType,
            formatType=formatType,
            fileSource=fileSource,
            versionId="none",
            partNumber="1",
            mileStone=mileStone,
        )
        return pt

    def getFilePathContentTypeTemplate(self, dataSetId, wfInstanceId=None, contentType=None, fileSource="archive", mileStone=None):
        _fp, _vt, _pt, cct = self.__getPathWorker(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            contentTypeBase=contentType,
            formatType="any",
            fileSource=fileSource,
            versionId="none",
            partNumber="1",
            mileStone=mileStone,
        )
        return cct

    def __getStandardPath(self, dataSetId, wfInstanceId=None, contentTypeBase=None, formatType=None, fileSource="archive", versionId="latest", partNumber="1", mileStone=None):
        fP = None
        try:
            fP, _vT, _pT, _ccT = self.__getPathWorker(
                dataSetId=dataSetId,
                wfInstanceId=wfInstanceId,
                contentTypeBase=contentTypeBase,
                formatType=formatType,
                fileSource=fileSource,
                versionId=versionId,
                partNumber=partNumber,
                mileStone=mileStone,
            )
        except Exception as e:
            logger.exception("+PathInfo.__getStandard() failing for %s id %s wf id %s with %r", fileSource, dataSetId, wfInstanceId, str(e))

        return fP

    def __getPathWorker(self, dataSetId, wfInstanceId=None, contentTypeBase=None, formatType=None, fileSource="archive", versionId="latest", partNumber="1", mileStone=None):
        """Return the path and templates corresponding to the input file typing arguments.

        Return:  <full file path>,<file path as a version template>,<file path as a partition template>
        """
        #
        try:
            if mileStone is not None:
                contentType = contentTypeBase + "-" + mileStone
            else:
                contentType = contentTypeBase
                #
            logger.debug(
                "+PathInfo.__getPathworker() file source %s for id %s wf id %s contentType %r formatType %r partNumber %r versionId %r",
                fileSource,
                dataSetId,
                wfInstanceId,
                contentType,
                formatType,
                partNumber,
                versionId,
            )
            dfRef = DataFileReference(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            if fileSource in ["archive", "wf-archive"]:
                dfRef.setDepositionDataSetId(dataSetId)
                dfRef.setStorageType("archive")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)
            elif fileSource in ["autogroup"]:
                dfRef.setDepositionDataSetId(dataSetId)
                dfRef.setStorageType("autogroup")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)
            elif fileSource in ["deposit"]:
                dfRef.setDepositionDataSetId(dataSetId)
                dfRef.setStorageType("deposit")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)
            elif fileSource in ["tempdep"]:
                dfRef.setDepositionDataSetId(dataSetId)
                dfRef.setStorageType("tempdep")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)
            elif fileSource == "wf-instance":
                dfRef.setDepositionDataSetId(dataSetId)
                dfRef.setWorkflowInstanceId(wfInstanceId)
                dfRef.setStorageType("wf-instance")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)
            elif fileSource in ["session", "wf-session"]:
                dfRef.setSessionPath(self.__sessionPath)
                dfRef.setSessionDataSetId(dataSetId)
                dfRef.setStorageType("session")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)

            elif fileSource in ["session-download"]:
                dfRef.setSessionPath(self.__sessionDownloadPath)
                dfRef.setSessionDataSetId(dataSetId)
                dfRef.setStorageType("session")
                dfRef.setContentTypeAndFormat(contentType, formatType)
                dfRef.setPartitionNumber(partNumber)
                dfRef.setVersionId(versionId)
            else:
                logger.debug("+PathInfo.__getPathworker() bad file source %s for id %s wf id %s contentType %r", fileSource, dataSetId, wfInstanceId, contentType)
                return None, None, None, None

            fP = None
            vT = None
            pT = None
            ctT = None
            if dfRef.isReferenceValid():
                fP = dfRef.getFilePathReference()
                dP = dfRef.getDirPathReference()
                pT = os.path.join(dP, dfRef.getPartitionNumberSearchTarget())
                vT = os.path.join(dP, dfRef.getVersionIdSearchTarget())
                ctT = os.path.join(dP, dfRef.getContentTypeSearchTarget())
                logger.debug("+PathInfo.__getPathworker() file path:                %s", fP)
                logger.debug("+PathInfo.__getPathworker() partition search path:    %s", pT)
                logger.debug("+PathInfo.__getPathworker() version search path:      %s", vT)
                logger.debug("+PathInfo.__getPathworker() content type search path: %s", ctT)
            else:
                dP = dfRef.getDirPathReference()
                try:
                    ctT = os.path.join(dP, dfRef.getContentTypeSearchTarget())
                except Exception as e:
                    ctT = None
                    logger.exception("+PathInfo.__getPathworker() failing with content type search template construction with %r", str(e))
                #
            return fP, vT, pT, ctT
        except Exception as e:
            logger.exception("Failing for source %s id %s wf id %s contentType %r with %r", fileSource, dataSetId, wfInstanceId, contentType, str(e))
        return None, None, None, None



"""
File:    mmCIFUtil.py
Author:  Zukang Feng
Update:  21-August-2012
Version: 001  Initial version

"""

__author__ = "Zukang Feng"
__email__ = "zfeng@rcsb.rutgers.edu"
__version__ = "V0.001"

import sys

from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer
from mmcif.io.PdbxReader import PdbxReader
from mmcif.io.PdbxWriter import PdbxWriter


class mmCIFUtil:
    """Using pdbx mmCIF utility to parse mmCIF file"""

    def __init__(self, verbose=False, log=sys.stderr, filePath=None):  # pylint: disable=unused-argument
        # self.__verbose = verbose
        self.__lfh = log
        self.__filePath = filePath
        self.__dataList = []
        self.__dataMap = {}
        self.__container = None
        self.__blockID = None
        self.__read()
        #

    def __read(self):
        if not self.__filePath:
            return
        #
        try:
            ifh = open(self.__filePath, "r")
            pRd = PdbxReader(ifh)
            pRd.read(self.__dataList)
            ifh.close()
            if self.__dataList:
                self.__container = self.__dataList[0]
                self.__blockID = self.__container.getName()
                idx = 0
                for container in self.__dataList:
                    self.__dataMap[container.getName()] = idx
                    idx += 1
                #
            #
        except Exception as e:
            self.__lfh.write("Read %s failed %s.\n" % (self.__filePath, str(e)))
        #

    def GetBlockID(self):
        """Return first block ID"""
        return self.__blockID

    def GetValueAndItemByBlock(self, blockName, catName):
        """Get category values and item names"""
        dList = []
        iList = []
        if blockName not in self.__dataMap:
            return dList, iList
        #
        catObj = self.__dataList[self.__dataMap[blockName]].getObj(catName)
        if not catObj:
            return dList, iList
        #
        iList = catObj.getAttributeList()
        rowList = catObj.getRowList()
        for row in rowList:
            tD = {}
            for idxIt, itName in enumerate(iList):
                if row[idxIt] != "?" and row[idxIt] != ".":
                    tD[itName] = row[idxIt]
            #
            if tD:
                dList.append(tD)
            #
        #
        return dList, iList

    def GetValueAndItem(self, catName):
        dList, iList = self.GetValueAndItemByBlock(self.__blockID, catName)
        return dList, iList

    def GetValue(self, catName):
        """Get category values based on category name 'catName'. The results are stored
        in a list of dictionaries with item name as key
        """
        dList, _iList = self.GetValueAndItemByBlock(self.__blockID, catName)
        return dList

    def GetSingleValue(self, catName, itemName):
        """Get the first value of item name 'itemName' from 'itemName' item in 'catName' category."""
        text = ""
        dlist = self.GetValue(catName)
        if dlist:
            if itemName in dlist[0]:
                text = dlist[0][itemName]
        return text
        #

    def UpdateSingleRowValue(self, catName, itemName, row, value):
        """Update value in single row"""
        catObj = self.__container.getObj(catName)
        if catObj is None:
            return
        #
        catObj.setValue(value, itemName, row)

    def UpdateMultipleRowsValue(self, catName, itemName, value):
        """Update value in multiple rows"""
        catObj = self.__container.getObj(catName)
        if catObj is None:
            return
        #
        rowNo = catObj.getRowCount()
        for row in range(0, rowNo):
            catObj.setValue(value, itemName, row)
        #

    def AddBlock(self, blockID):
        """Add Data Block"""
        self.__container = DataContainer(blockID)
        self.__blockID = blockID
        self.__dataMap[blockID] = len(self.__dataList)
        self.__dataList.append(self.__container)

    def AddCategory(self, categoryID, items):
        """Add Category"""
        category = DataCategory(categoryID)
        for item in items:
            category.appendAttribute(item)
        #
        self.__container.append(category)

    def RemoveCategory(self, categoryID):
        return self.__container.remove(categoryID)

    def InsertData(self, categoryID, dataList):
        """"""
        catObj = self.__container.getObj(categoryID)
        if catObj is None:
            return
        #
        for data in dataList:
            catObj.append(data)
        #

    def WriteCif(self, outputFilePath=None):
        """Write out cif file"""
        if not outputFilePath:
            return
        #
        ofh = open(outputFilePath, "w")
        pdbxW = PdbxWriter(ofh)
        pdbxW.write(self.__dataList)
        ofh.close()

    def GetCategories(self):
        return self.__container.getObjNameList()

    def GetAttributes(self, category):
        return self.__container.getObj(category).getAttributeList()

    def category_as_dict(self, category, block=None):
        if block is None:
            block = self.__blockID
        values, attributes = self.GetValueAndItemByBlock(block, category)
        data = [[x[y] if y in x else None for y in attributes] for x in values]
        return {category: {"Items": attributes, "Values": data}}

    def block_as_dict(self, block=None):
        if block is None:
            block = self.__blockID
        data = {}
        for category in self.GetCategories():
            data.update(self.category_as_dict(category, block=block))
        return data
